from ..models import DoctrineReport, FittingReport
from .testdata.load_eveuniverse import load_eveuniverse
from .utils import CorphandoutsTestCase
from .utils.fitting import create_sabre_fitting


class TestTasks(CorphandoutsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_create_basic_report(self):
        fitting = create_sabre_fitting()

        doctrine_report = DoctrineReport.objects.create(
            name="Test doctrine",
            corporation=self.corporation_audit,
            location_id=self.location.location_id,
            corporation_hangar_division=1,
        )

        FittingReport.objects.create(
            doctrine=doctrine_report,
            fit=fitting,
        )



from corptools.models import CorpAsset, CorporationAudit, EveItemType, EveLocation

from allianceauth.eveonline.models import EveCorporationInfo

CORPORATION_ID = 109299958
SHIP_ITEM_ID = 1042868888363


def _create_corptools_corporation():
    """Creates the C C P corporation in corptools"""
    corporation_info = EveCorporationInfo(
        corporation_id=CORPORATION_ID
    ).update_corporation()
    CorporationAudit.objects.create(corporation=corporation_info)


def corptools_create_fitted_sabre(
    location: EveLocation, corporation: CorporationAudit
) -> CorpAsset:
    """
    Adds a fitted sabre in corp hangar 1
    Returns the sabre Asset
    """
    corporation_id = corporation.corporation.corporation_id
    sabre = __create_corptools_corp_asset(
        True,
        SHIP_ITEM_ID,
        "CorpSAG1",
        location.location_id,
        1,
        22456,
        corporation_id,
        "test sabre",
    )
    sabre_location = EveLocation.objects.create(
        location_name="sabre location", location_id=SHIP_ITEM_ID
    )
    location_id = sabre_location.location_id
    for i in range(6):
        __create_corptools_corp_asset(
            True,
            100 + i,
            location_flag=f"HiSlot{i}",
            location_id=location_id,
            quantity=1,
            type_id=2873,
            corporation_id=corporation_id,
        )  # 125mm AC
    __create_corptools_corp_asset(
        True,
        110,
        location_flag="HiSlot6",
        location_id=location_id,
        quantity=1,
        type_id=22782,
        corporation_id=corporation_id,
    )  # Sphere launcher
    __create_corptools_corp_asset(
        True,
        120,
        location_flag="HiSlot7",
        location_id=location_id,
        quantity=1,
        type_id=20565,
        corporation_id=corporation_id,
    )  # Cloak

    for i in range(2):
        __create_corptools_corp_asset(
            True,
            200 + i,
            location_flag=f"MedSlot{i}",
            location_id=location_id,
            quantity=1,
            type_id=8517,
            corporation_id=corporation_id,
        )  # Compact MSE
    __create_corptools_corp_asset(
        True,
        210,
        location_flag="MedSlot2",
        location_id=location_id,
        quantity=1,
        type_id=35658,
        corporation_id=corporation_id,
    )  # 5mn
    __create_corptools_corp_asset(
        True,
        220,
        location_flag="MedSlot3",
        location_id=location_id,
        quantity=1,
        type_id=40758,
        corporation_id=corporation_id,
    )  # Scram

    for i in range(2):
        __create_corptools_corp_asset(
            True,
            300 + i,
            location_flag=f"LoSlot{i}",
            location_id=location_id,
            quantity=1,
            type_id=2605,
            corporation_id=corporation_id,
        )  # Nanos
        __create_corptools_corp_asset(
            True,
            400 + i,
            location_flag=f"RigSlot{i}",
            location_id=location_id,
            quantity=1,
            type_id=31165,
            corporation_id=corporation_id,
        )  # Hypers

    __create_corptools_corp_asset(
        False,
        500,
        location_flag="Cargo",
        location_id=location_id,
        quantity=2000,
        type_id=12608,
        corporation_id=corporation_id,
    )  # Hail
    __create_corptools_corp_asset(
        False,
        501,
        location_flag="Cargo",
        location_id=location_id,
        quantity=1000,
        type_id=12625,
        corporation_id=corporation_id,
    )  # Barrage

    return sabre


def __create_corptools_corp_asset(
    singleton: bool,
    item_id: int,
    location_flag: str,
    location_id: int,
    quantity: int,
    type_id: int,
    corporation_id: int,
    item_name: str = None,
) -> CorpAsset:
    return CorpAsset.objects.create(
        singleton=singleton,
        item_id=item_id,
        location_flag=location_flag,
        location_id=location_id,
        location_type="Station",
        quantity=quantity,
        type_id=type_id,
        type_name=EveItemType.objects.get_or_create_from_esi(type_id)[0],
        location_name=EveLocation.objects.get(location_id=location_id),
        corporation=CorporationAudit.objects.get(
            corporation__corporation_id=corporation_id
        ),
        name=item_name,
    )



from corptools.models import CorpAsset, CorporationAudit, EveLocation

from django.db.models import Q, QuerySet

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def get_assets_corporation_division(
    corporation: CorporationAudit, location: EveLocation, hangar_division: int
) -> QuerySet[CorpAsset]:
    """Returns a queryset of corporation assets in the specified hangar and location"""

    corp_assets = CorpAsset.objects.filter(corporation=corporation)
    asset_locations = corp_assets.filter(location_name_id=location.location_id)
    corp_assets = corp_assets.filter(
        Q(location_name_id=location.location_id)
        | Q(location_id__in=asset_locations.values_list("item_id"))
        | Q(location_id=location.location_id)
    )

    return corp_assets.filter(location_flag=f"CorpSAG{hangar_division}")


def get_ship_fit(ship_item_id: int) -> QuerySet[CorpAsset]:
    """Returns all known items linked to this ship id"""

    return CorpAsset.objects.filter(location_id=ship_item_id)



"""Models."""

from corptools.models import CorporationAudit, EveLocation
from fittings.models import Fitting

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveType


class General(models.Model):
    """A metamodel for app permissions."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and view corporation handouts"),
            ("manager", "Can view handouts for all corporations"),
        )


class DoctrineReport(models.Model):
    """Defines a group of fittings to be linked together in reports"""

    name = models.CharField(max_length=50, help_text=_("Doctrine name"))
    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        help_text=_("Corporation to get the assets from"),
    )
    location = models.ForeignKey(
        EveLocation,
        on_delete=models.CASCADE,
        help_text=_("Where the doctrine should be located"),
    )
    corporation_hangar_division = models.IntegerField(
        default=1, validators=[MinValueValidator(1), MaxValueValidator(7)]
    )

    def __str__(self):
        return f"{self.name} / {self.corporation.corporation.corporation_name} / {self.location.location_name}"

    @property
    def first_ship_type_id(self) -> int:
        """Returns the type if of the first ship in the ship list"""
        fittings = self.fittings.all()
        if fittings.exists():
            return fittings[0].fit.ship_type_type_id

    @property
    def ship_types_ids(self) -> list[int]:
        """Return all ship type ids in the doctrine"""
        return self.fittings.values_list("fit__ship_type__id", flat=True)

    @property
    def count_corrections(self) -> int:
        """Return the number of fits to correct known for this corporation"""
        return self.fittings.values("fit_to_correct").count()


class FittingReport(models.Model):
    """Defines how a fit should be found in the corporation hangars"""

    doctrine = models.ForeignKey(
        DoctrineReport, on_delete=models.CASCADE, related_name="fittings"
    )
    fit = models.ForeignKey(Fitting, on_delete=models.CASCADE)
    expected_amount = models.PositiveIntegerField(
        null=True,
        blank=True,
        help_text=_(
            "Amount of fits expected in the corp hangar. Set to 0 to ignore this fit"
        ),
    )
    regex = models.CharField(
        max_length=40,
        null=True,
        blank=True,
        help_text=_("Optional regex to get ships from"),
    )

    ok_ships = models.PositiveIntegerField(
        default=0, help_text=_("Current amount of ships fitted correctly")
    )

    @property
    def count_errors(self) -> int:
        """Return the number of fits to fix"""
        return self.fits_to_correct.count()

    def __str__(self):
        return f"{self.fit} {self.doctrine.name}"


class FittingToCorrect(models.Model):
    """
    Represents a fit that needs to be fixed
    """

    item_name = models.CharField(max_length=40, help_text=_("Name of the ship"))
    item_id = models.BigIntegerField(help_text=_("Item id if the ship§"))
    fit = models.ForeignKey(
        FittingReport,
        on_delete=models.CASCADE,
        related_name="fits_to_correct",
        related_query_name="fit_to_correct",
    )

    def __str__(self):
        return f"{self.fit.fit.ship_type.name} / {self.item_name}"

    @property
    def id(self):
        """Return primary key id"""
        return self.item_id

    def get_correction_icons(self) -> list[int]:
        """Return a list of type ids for icons representing the categories that need corrections"""
        icon_type_ids = []
        if self._category_has_error("Charge"):
            icon_type_ids.append(230)  # Antimatter M
        if self._category_has_error("Module"):
            icon_type_ids.append(496)  # 800mm cannon I
        if self._category_has_error("Implant"):
            icon_type_ids.append(28680)  # Synth mindflood
        return icon_type_ids

    def _category_has_error(self, category_name: str) -> bool:
        """Return true if among the fit corrections one of them is part of the eve category with this name"""
        return self.corrections.filter(
            eve_type__eve_group__eve_category__name=category_name
        ).exists()


class FittingCorrection(models.Model):
    """
    Represents a correction to do on a fit
    """

    class CorrectionType(models.TextChoices):
        """Cargo of fitting issue"""

        CARGO = "CA", "Cargo"
        FITTING = "FI", "Fitting"

        @classmethod
        def from_value_to_label(cls, value: str | None) -> str:
            """Return the label from one of the possible values"""
            match value:
                case "CA":
                    return _("Cargo")
                case "FI":
                    return _("Fitting")
                case _:
                    return _("Unknown")

    fit_to_correct = models.ForeignKey(
        FittingToCorrect,
        on_delete=models.CASCADE,
        related_name="corrections",
        related_query_name="correction",
    )
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE)
    correction = models.IntegerField(
        help_text=_("Amount to remove/add to the fit to match the template")
    )
    correction_type = models.CharField(max_length=2, choices=CorrectionType.choices)

    def __str__(self):
        return f"({self.fit_to_correct}) / {self.eve_type} / {self.correction}"



from corptools.models import CorporationAudit, EveLocation

from django.test import TestCase

from allianceauth.eveonline.models import EveCorporationInfo

from corphandouts.models import DoctrineReport, FittingReport

from .fitting import create_sabre_fitting

LOCATION_ID = 1


class CorphandoutsTestCase(TestCase):
    """Test case preloading the data needed for this app"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        location, corporation_audit = preload_corptools_data()
        cls.location = location
        cls.corporation_audit = corporation_audit

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.location.delete()
        cls.corporation_audit.delete()


def preload_corptools_data() -> tuple[EveLocation, CorporationAudit]:
    """Basic preload function"""
    location = EveLocation.objects.create(
        location_id=LOCATION_ID, location_name="TestStationLocation"
    )
    corp = EveCorporationInfo.objects.get_or_create(
        corporation_id=123,
        corporation_name="corporation.name1",
        corporation_ticker="ABC",
        member_count=1,
        ceo_id=1,
    )[0]
    corpaudit = CorporationAudit.objects.create(corporation=corp)

    return location, corpaudit


def create_basic_doctrine_report(
    corporation_audit: CorporationAudit, location: EveLocation
) -> DoctrineReport:
    """Creates a simple doctrine report"""
    saved_fitting = create_sabre_fitting()

    doctrine_report = DoctrineReport.objects.create(
        name="Test doctrine",
        corporation=corporation_audit,
        location_id=location.location_id,
        corporation_hangar_division=1,
    )

    FittingReport.objects.create(
        doctrine=doctrine_report,
        fit=saved_fitting,
    )

    return doctrine_report



from django.core.management.base import BaseCommand

from corphandouts.tasks import update_all_doctrine_reports


class Command(BaseCommand):

    def handle(self, *args, **options):
        update_all_doctrine_reports()



"""Views."""

from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import get_object_or_404, render
from django.views.decorators.cache import cache_page

from allianceauth.services.hooks import get_extension_logger

from corphandouts.models import DoctrineReport, FittingReport, FittingToCorrect

logger = get_extension_logger(__name__)


def add_common_context(context: dict = None) -> dict:
    """Enhance the templates context with context that should be added to every page"""
    if context is None:
        context = {}

    if basic_title := context.get("page_title"):
        context["page_title"] = f"{basic_title} - Corporation Handouts"
    else:
        context["page_title"] = "Corporation Handouts"

    return context


@login_required
@permission_required("corphandouts.basic_access")
def index(request):
    """Render index view."""
    doctrines = DoctrineReport.objects.all()
    return render(
        request,
        "corphandouts/index.html",
        add_common_context(
            {
                "doctrines": doctrines,
            }
        ),
    )


@login_required
@permission_required("corphandouts.basic_access")
def doctrine_fitting_reports(request, doctrine_id: int):
    """Shows all fitting reports for this doctrine"""
    logger.info("Trying to show fitting reports for doctrine id %d", doctrine_id)
    doctrine_report = get_object_or_404(DoctrineReport, pk=doctrine_id)

    fitting_reports = doctrine_report.fittings.all()
    logger.debug(fitting_reports)

    return render(
        request,
        "corphandouts/doctrine.html",
        add_common_context(
            {
                "doctrine_report": doctrine_report,
                "fitting_reports": fitting_reports,
            }
        ),
    )


@login_required
@permission_required("corphandouts.basic_access")
def doctrine_fit_reports_all(request, doctrine_id: int):
    """Shows all fittings with errors for this doctrine"""
    logger.info("Trying to show all fitting errors for doctrine id %d", doctrine_id)
    doctrine_report = get_object_or_404(DoctrineReport, pk=doctrine_id)
    logger.debug(doctrine_report)

    fittings = []
    for fitting_report in doctrine_report.fittings.all():
        fittings.extend(fitting_report.fits_to_correct.all())
    logger.debug(fittings)

    return render(
        request,
        "corphandouts/fittings.html",
        add_common_context({"doctrine_id": doctrine_id, "fittings": fittings}),
    )


@login_required
@permission_required("corphandouts.basic_access")
def fitting_reports(request, doctrine_id: int, fitting_report_id: int):
    """Doctrine report view"""
    fitting_report = get_object_or_404(
        FittingReport, doctrine_id=doctrine_id, pk=fitting_report_id
    )
    logger.info(fitting_report)

    fittings = fitting_report.fits_to_correct.all()
    logger.debug(fittings)

    return render(
        request,
        "corphandouts/fittings.html",
        add_common_context(
            {"doctrine_id": doctrine_id, "fittings": fittings},
        ),
    )


@login_required
@permission_required("corphandouts.basic_access")
def fitting(request, fitting_id: int):
    """Fitting report"""
    fitting = get_object_or_404(FittingToCorrect, pk=fitting_id)

    context = {
        "fitting": fitting,
    }

    logger.info(fitting_id)

    if request.GET.get("new_page"):
        context["title"] = "title"
        context["content_file"] = "corphandouts/partials/fit_corrections.html"
        return render(
            request,
            "corphandouts/modals/generic_modal_page.html",
            add_common_context(context),
        )
    return render(
        request, "corphandouts/modals/fit_corrections.html", add_common_context(context)
    )


@cache_page(3600)
def modal_loader_body(request):
    """Draw the loader body. Useful for showing a spinner while loading a modal."""
    return render(request, "corphandouts/modals/loader_body.html")



from django import template

from corphandouts.models import FittingCorrection

register = template.Library()


@register.filter
def correction_type(correction_type_value: str) -> str:
    """Filter to display a correction type correctly"""
    return FittingCorrection.CorrectionType.from_value_to_label(correction_type_value)



from pathlib import Path

_current_dir = Path(__file__).parent
_FILENAME_EVEUNIVERSE_TESTDATA = "eveuniverse.json"


def test_data_filename() -> str:
    return str(_current_dir / _FILENAME_EVEUNIVERSE_TESTDATA)



"""Tasks."""

from collections import Counter

from celery import group, shared_task
from corptools.models import CorpAsset

from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger

from corphandouts.corptools import get_assets_corporation_division, get_ship_fit
from corphandouts.models import (
    DoctrineReport,
    FittingCorrection,
    FittingReport,
    FittingToCorrect,
)

logger = get_extension_logger(__name__)


@shared_task()
def update_all_doctrine_reports():
    """Runs the update command on all doctrines"""
    doctrines = DoctrineReport.objects.all()
    logger.info("Running the update command on %d doctrines", doctrines.count())

    tasks = [update_doctrine_report.si(doctrine.id) for doctrine in doctrines]
    group(tasks).delay()


@shared_task()
def update_doctrine_report(doctrine_report_id: int):
    """
    Goes through corporation assets to check what fits need to be updated
    """
    logger.info("Updating doctrine report id %d", doctrine_report_id)
    doctrine_report = DoctrineReport.objects.get(id=doctrine_report_id)
    assets_in_corporation_division = get_assets_corporation_division(
        doctrine_report.corporation,
        doctrine_report.location,
        doctrine_report.corporation_hangar_division,
    )
    logger.debug(
        "%d assets fetched in the corporation division",
        assets_in_corporation_division.count(),
    )

    tasks = []
    for fitting_to_check in doctrine_report.fittings.all():
        fit_ship_type_id = fitting_to_check.fit.ship_type_type_id
        item_ids_to_check = assets_in_corporation_division.filter(
            type_id=fit_ship_type_id
        ).values_list("item_id", flat=True)
        logger.debug(
            "%d item with type id %d found %s",
            item_ids_to_check.count(),
            fit_ship_type_id,
            list(item_ids_to_check),
        )
        tasks.append(
            check_doctrine_fit.si(fitting_to_check.id, list(item_ids_to_check))
        )

    group(tasks).delay()


@shared_task()
def check_doctrine_fit(fitting_report_id: int, ships_item_ids: list[int]):
    """
    Receives a list of item_ids that are in the hangar appropriate for this fitting and checks which of them are fitted correctly
    """
    logger.info("Updating fitting report id %d", fitting_report_id)
    logger.debug(ships_item_ids)
    fitting_report = FittingReport.objects.get(id=fitting_report_id)

    # Deletes existing corrections if any to overwrite
    FittingToCorrect.objects.filter(fit=fitting_report).delete()

    expected_fitting_items = fitting_report.fit.items.all()
    expected_fitting_counter = Counter(
        expected_fitting_items.filter(
            flag__regex=r"^(Hi|Med|Lo|Rig|SubSystem)Slot\d$"
        ).values_list("type_id", flat=True)
    )
    expected_cargo_counter = Counter(
        {
            eve_type_id: count
            for (eve_type_id, count) in expected_fitting_items.filter(
                flag="Cargo"
            ).values_list("type_id", "quantity")
        }
    )
    logger.debug(expected_fitting_counter)
    logger.debug(expected_cargo_counter)

    fit_ok_count = 0
    fits_to_fix = []

    for ships_item_id in ships_item_ids:
        logger.debug("Checking item id %d", ships_item_id)

        current_ship_items = get_ship_fit(ships_item_id)
        current_fitting_counter = Counter(
            current_ship_items.filter(
                location_flag__regex=r"^(Hi|Med|Lo|Rig|SubSystem)Slot\d$"
            ).values_list("type_id", flat=True)
        )
        # TODO Check the "AutoFit" flag items to move in cargo
        current_cargo_counter = Counter(
            {
                type_id: count
                for (type_id, count) in current_ship_items.filter(
                    location_flag="Cargo"
                ).values_list("type_id", "quantity")
            }
        )
        logger.debug(current_cargo_counter)
        current_fitting_counter.subtract(expected_fitting_counter)
        current_cargo_counter.subtract(expected_cargo_counter)
        logger.debug(current_fitting_counter)
        logger.debug(current_cargo_counter)

        if all(current_fitting_counter.values()) and all(
            current_cargo_counter.values()
        ):
            logger.debug("%d fit is correct", ships_item_id)
            fit_ok_count += 1

        else:
            logger.info("%d fit is incorrect", ships_item_id)
            fits_to_fix.append(
                (ships_item_id, current_fitting_counter, current_cargo_counter)
            )

    logger.debug("Ok ships: %d", fit_ok_count)
    fitting_report.ok_ships = fit_ok_count
    fitting_report.save()

    if not fits_to_fix:
        logger.info("All fits fitted correctly, exiting")
        return

    # TODO split in tasks
    for item_type_id, fitting_difference, cargo_difference in fits_to_fix:
        corp_asset = CorpAsset.objects.get(item_id=item_type_id)
        fitting_to_correct = FittingToCorrect.objects.create(
            fit=fitting_report,
            item_id=corp_asset.item_id,
            item_name=corp_asset.name,
        )
        corrections = []
        for type_id, amount in fitting_difference.items():
            if amount:
                amount = -amount
                eve_type, _ = EveType.objects.get_or_create_esi(id=type_id)
                corrections.append(
                    FittingCorrection(
                        fit_to_correct=fitting_to_correct,
                        eve_type=eve_type,
                        correction=amount,
                        correction_type=FittingCorrection.CorrectionType.FITTING,
                    )
                )
        for type_id, amount in cargo_difference.items():
            if amount:
                amount = -amount
                eve_type, _ = EveType.objects.get_or_create_esi(id=type_id)
                corrections.append(
                    FittingCorrection(
                        fit_to_correct=fitting_to_correct,
                        eve_type=eve_type,
                        correction=amount,
                        correction_type=FittingCorrection.CorrectionType.CARGO,
                    )
                )

        logger.debug(corrections)
        FittingCorrection.objects.bulk_create(corrections)



from django.test import TestCase
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        moon_goo_id_range = [i for i in range(16633, 16654)]
        moon_goo_id_range.remove(16645)  # why does this one not exists? #CCPLZ
        testdata_spec = [
            ModelSpec(
                "EveType",
                ids=[
                    22456,  # Sabre
                    2873,  # 125mm AC II
                    22782,  # Sphere launcher
                    20565,  # Cloak
                    8517,  # Compact MSE
                    35658,  # 5MN
                    40758,  # Scram II
                    2605,  # Nanos II
                    31165,  # Small hyper II
                    12608,  # Hail S
                    12625,  # Barrage S
                    22778,  # Bubble
                ],
            )
        ]
        create_testdata(testdata_spec, test_data_filename())



"""ESI interactions"""

from corptools.task_helpers.corp_helpers import get_corp_token

from esi.clients import EsiClientProvider

from allianceauth.services.hooks import get_extension_logger

from . import __version__

logger = get_extension_logger(__name__)
esi = EsiClientProvider(app_info_text=f"aa-corp-handouts v{__version__}")


def get_corporation_asset_names(
    corporation_id: int, item_ids: list[int]
) -> list[[int, int]]:
    """
    Return names for a set of item ids, which you can get from corporation assets endpoint.
    """
    logger.info("Updating names for corp id %d", corporation_id)
    logger.debug(item_ids)

    req_scopes = ["esi-assets.read_corporation_assets.v1"]
    req_roles = ["Director"]
    token = get_corp_token(corporation_id, req_scopes, req_roles)

    return esi.client.Assets.post_corporations_corporation_id_assets_names(
        corporation_id=corporation_id,
        item_ids=item_ids,
        token=token.valid_access_token(),
    ).result()



from corptools.models import EveItemType

from ..corptools import get_ship_fit
from ..models import FittingCorrection, FittingToCorrect
from ..tasks import update_all_doctrine_reports
from .testdata.load_eveuniverse import load_eveuniverse
from .utils import CorphandoutsTestCase, create_basic_doctrine_report
from .utils.corptools import corptools_create_fitted_sabre


class TestTasks(CorphandoutsTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_update_doctrine_report_no_errors(self):
        create_basic_doctrine_report(self.corporation_audit, self.location)
        corptools_create_fitted_sabre(self.location, self.corporation_audit)

        update_all_doctrine_reports()

        self.assertEqual(0, FittingCorrection.objects.count())

    def test_update_doctrine_report_fitting_error(self):
        create_basic_doctrine_report(self.corporation_audit, self.location)
        sabre = corptools_create_fitted_sabre(self.location, self.corporation_audit)
        sabre_fitting = get_ship_fit(sabre.item_id)

        slot_0_125mm = sabre_fitting.get(location_flag="HiSlot0", type_id=2873)
        slot_0_125mm.type_id = 484
        slot_0_125mm.type_name = EveItemType.objects.get_or_create_from_esi(484)[0]
        slot_0_125mm.save()

        update_all_doctrine_reports()

        self.assertEqual(2, FittingCorrection.objects.count())

    def test_ships_are_deleted_on_updates(self):
        doctrine_report = create_basic_doctrine_report(
            self.corporation_audit, self.location
        )
        fitting_report = doctrine_report.fittings.all()[0]
        FittingToCorrect.objects.create(
            item_name="Test to delete",
            item_id=1,
            fit=fitting_report,
        )

        update_all_doctrine_reports()

        self.assertEqual(
            0,
            FittingToCorrect.objects.filter(
                item_name="Test to delete", item_id=1
            ).count(),
        )



from fittings.models import Fitting, FittingItem

from eveuniverse.models import EveType


def create_sabre_fitting() -> Fitting:
    fit = Fitting.objects.create(
        name="Test fit",
        ship_type=EveType.objects.get(id=22456),
        ship_type_type_id=22456,
    )

    for i in range(6):
        __add_item_to_fit(fit, f"HiSlot{i}", type_id=2873)  # 125mm II
    __add_item_to_fit(fit, "HiSlot6", 22782)  # Sphere launcher
    __add_item_to_fit(fit, "HiSlot7", 20565)  # Cloak

    for i in range(2):
        __add_item_to_fit(fit, f"MedSlot{i}", type_id=8517)  # MSE
    __add_item_to_fit(fit, "MedSlot2", type_id=35658)  # 5mn
    __add_item_to_fit(fit, "MedSlot3", type_id=40758)  # scram

    for i in range(2):
        __add_item_to_fit(fit, f"LoSlot{i}", type_id=2605)  # Nanos
        __add_item_to_fit(fit, f"RigSlot{i}", type_id=31165)  # Hypers

    __add_item_to_fit(fit, "Cargo", type_id=12608, quantity=2000)  # Hail S
    __add_item_to_fit(fit, "Cargo", type_id=12625, quantity=1000)  # Barrage S

    return fit


def __add_item_to_fit(
    fit: Fitting, flag: str, type_id: int, quantity: int | None = None
):
    FittingItem.objects.create(
        fit=fit,
        flag=flag,
        quantity=quantity or 1,
        type_fk=EveType.objects.get(id=type_id),
        type_id=type_id,
    )



from corphandouts.corptools import get_assets_corporation_division, get_ship_fit

from .utils import CorphandoutsTestCase
from .utils.corptools import corptools_create_fitted_sabre


class CorptoolsTest(CorphandoutsTestCase):

    def test_get_assets_corporation_division(self):
        sabre_asset = corptools_create_fitted_sabre(
            self.location, self.corporation_audit
        )

        assets_in_corp_division = get_assets_corporation_division(
            self.corporation_audit, self.location, 1
        )

        self.assertIn(sabre_asset, assets_in_corp_division)
        self.assertEqual(1, assets_in_corp_division.count())

    def test_get_ship_fit(self):
        sabre_asset = corptools_create_fitted_sabre(
            self.location, self.corporation_audit
        )

        sabre_fit = get_ship_fit(sabre_asset.item_id)

        self.assertEqual(18, sabre_fit.count())
        self.assertEqual(
            8, sabre_fit.filter(location_flag__regex=r"^HiSlot\d$").count()
        )
        self.assertEqual(
            4, sabre_fit.filter(location_flag__regex=r"^MedSlot\d$").count()
        )
        self.assertEqual(
            2, sabre_fit.filter(location_flag__regex=r"^LoSlot\d$").count()
        )
        self.assertEqual(
            2, sabre_fit.filter(location_flag__regex=r"^RigSlot\d$").count()
        )
        self.assertEqual(2000, sabre_fit.get(type_id=12608).quantity)
        self.assertEqual(1000, sabre_fit.get(type_id=12625).quantity)



# Generated by Django 4.2.16 on 2025-04-10 19:17

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("corptools", "0120_corptoolsconfiguration_aggregate_lookback"),
        (
            "fittings",
            "0016_remove_dogmaattribute_type_remove_dogmaeffect_type_and_more",
        ),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "basic_access",
                        "Can access this app and view corporation handouts",
                    ),
                    ("manager", "Can view handouts for all corporations"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="DoctrineReport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(help_text="Doctrine name", max_length=50)),
                (
                    "corporation_hangar_division",
                    models.IntegerField(
                        default=1,
                        validators=[
                            django.core.validators.MinValueValidator(1),
                            django.core.validators.MaxValueValidator(7),
                        ],
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        help_text="Corporation to get the assets from",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="corptools.corporationaudit",
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        help_text="Where the doctrine should be located",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="corptools.evelocation",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FittingReport",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "expected_amount",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Amount of fits expected in the corp hangar. Set to 0 to ignore this fit",
                        null=True,
                    ),
                ),
                (
                    "regex",
                    models.CharField(
                        blank=True,
                        help_text="Optional regex to get ships from",
                        max_length=40,
                        null=True,
                    ),
                ),
                (
                    "ok_ships",
                    models.PositiveIntegerField(
                        default=0, help_text="Current amount of ships fitted correctly"
                    ),
                ),
                (
                    "doctrine",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fittings",
                        to="corphandouts.doctrinereport",
                    ),
                ),
                (
                    "fit",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="fittings.fitting",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FittingToCorrect",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "item_name",
                    models.CharField(help_text="Name of the ship", max_length=40),
                ),
                ("item_id", models.BigIntegerField(help_text="Item id if the ship§")),
                (
                    "fit",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fits_to_correct",
                        related_query_name="fit_to_correct",
                        to="corphandouts.fittingreport",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FittingCorrection",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "correction",
                    models.IntegerField(
                        help_text="Amount to remove/add to the fit to match the template"
                    ),
                ),
                (
                    "correction_type",
                    models.CharField(
                        choices=[("CA", "Cargo"), ("FI", "Fitting")], max_length=2
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "fit_to_correct",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="corrections",
                        related_query_name="correction",
                        to="corphandouts.fittingtocorrect",
                    ),
                ),
            ],
        ),
    ]



"""Admin site."""

from corptools.models import CorpAsset, EveLocation

from django import forms
from django.contrib import admin

from corphandouts.models import DoctrineReport, FittingReport


class DoctrineReportForm(forms.ModelForm):
    class Meta:
        model = DoctrineReport
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        asset_locations_names_queryset = (
            CorpAsset.objects.filter(location_name__isnull=False)
            .values_list("location_name")
            .distinct()
        )
        asset_locations_queryset = EveLocation.objects.filter(
            location_id__in=asset_locations_names_queryset, managed=False
        ).order_by("location_name")
        # TODO filter out metenoxes
        self.fields["location"].queryset = asset_locations_queryset


class FitInline(admin.TabularInline):
    model = FittingReport
    readonly_fields = ["ok_ships"]


@admin.register(DoctrineReport)
class DoctrineReportAdmin(admin.ModelAdmin):
    form = DoctrineReportForm
    inlines = (FitInline,)



"""Routes."""

from django.urls import path

from . import views

app_name = "corphandouts"

urlpatterns = [
    path("", views.index, name="index"),
    path(
        "doctrine/<int:doctrine_id>",
        views.doctrine_fitting_reports,
        name="doctrine_report",
    ),
    path(
        "doctrine/<int:doctrine_id>/fits",
        views.doctrine_fit_reports_all,
        name="doctrine_fit_report_all",
    ),
    path(
        "doctrine/<int:doctrine_id>/fits/<int:fitting_report_id>",
        views.fitting_reports,
        name="doctrine_fit_report",
    ),
    path("fitting/<int:fitting_id>", views.fitting, name="fitting"),
    path("modal_loader_body", views.modal_loader_body, name="modal_loader_body"),
]
